#|
This file is part of Ordo.

Copyright (C) 2025 Ray Miller

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, version 3.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.
|#

(define-module (ordo play)
  #:use-module (ordo connection)
  #:use-module (ordo context)
  #:use-module (ordo handler)
  #:use-module (ordo inventory)
  #:use-module (ordo logger)
  #:use-module (ordo task)
  #:use-module (ordo util flatten)
  #:use-module (srfi srfi-9)
  #:use-module (srfi srfi-26)
  #:use-module (srfi srfi-69)
  #:export (play
            play?
            play-host
            play-sudo?
            play-sudo-user
            play-sudo-password
            play-vars
            play-tasks
            play-handlers
            run-play))

(define-record-type <play>
  (make-play name host sudo? sudo-user sudo-password vars tasks handlers)
  play?
  (name play-name)
  (host play-host)
  (sudo? play-sudo?)
  (sudo-user play-sudo-user)
  (sudo-password play-sudo-password)
  (vars play-vars)
  (tasks play-tasks)
  (handlers play-handlers))

(define* (play #:key name host (sudo? #f) (sudo-user #f) (sudo-password #f) (vars '()) (tasks '()) (handlers '()))
  (make-play name host sudo? sudo-user sudo-password (alist->hash-table vars) tasks handlers))

(define (run-play p)
  (log-msg 'NOTICE "Running play: " (play-name p))
  (parameterize ((*play-handlers* (map handler-name (play-handlers p)))
                 (*play-vars* (play-vars p)))
    (let ((hosts (resolve-hosts (*inventory*) (play-host p))))
      (if (null? hosts)
          (log-msg 'WARN "No hosts matched: " (play-host p))
          (for-each (cut run-host-play p <>) hosts)))))

(define (run-host-play p h)
  (log-msg 'NOTICE "Running play on host: " (host-name h))
  (parameterize ((*host-vars* (host-vars h))
                 (*play-handlers* (play-handlers p))
                 (*play-triggers* (make-bitvector (length (play-handlers p)) #f)))
    (call-with-connection
     (host-connection h)
     (lambda (conn)
       (for-each (cut run-task <> conn) (play-tasks p))
       (for-each (lambda (h i)
                   (when (bitvector-bit-set? (*play-triggers*) i)
                     (run-handler h conn)))
                 (play-handlers p) (iota (length (play-handlers p)))))
     #:sudo? (play-sudo? p)
     #:sudo-user (play-sudo-user p)
     #:sudo-password (play-sudo-password p))))
